"""
Concurrence
Copyright (C) 2005 Daniel Bickett

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
"""

# Twisted Imports
from twisted.protocols.basic import LineReceiver
from twisted.internet.protocol import Factory
from twisted.internet import reactor

# Standard Modules
import base64
import md5
import operator
import os

# Custom Modules
import ConcurrenceProtocol as codes

#
#  Misc Functions
#

def strIns( string , pos , text ):
	# insert <text> into <string> at <pos>
	try:
		s = list( string )
		s.insert( pos , text )
		return ''.join( s )
	except:
		return False

def strDel( string , pos , sliceLen ):
	# remove <sliceLen> characers from <string> starting at <pos>
	try:
		s = list( string )
		del s[ pos : pos + sliceLen ]
		return ''.join( s )
	except:
		return False

# Thanks to Gordon McMillan; http://tinyurl.com/933vq
def splitEqual( string , pieces ):
	length = ( len( string ) + pieces - 1 ) / pieces
	arg1 = [ string ] * pieces
	arg2 = range( 0 , pieces * length , length )
	arg3 = range( length , ( pieces + 1 ) * length , length )
	return map( operator.getslice , arg1 , arg2 , arg3 )

# Protocol Resource
# - Central dictionary where a handle on every single protocol is placed
protocols = {}
# Files
# - Central dictionary where the server's files are stored
# Key: b64 encoded filename
# Args: 'name' : file name, 'data' : file contents, 'lines' : user line table
files = {}

def distSend( text , id = None ):
	"""Sends a message to everyone but optional <id>"""
	for key in protocols.iterkeys():
		if not id or ( id and key != id ):
			protocols[key].sendLine( text )
	return True

def encode( text ):
	"""Shortens the process of base64 encoding and stripping the result of whitespace"""
	return base64.encodestring( text ).strip()

#
#  Twisted Code
#

class ConcurrenceServerProtocol( LineReceiver ):
	
	def __init__( self ):
		# indication of what files this client is viewing/modifying
		self.files = []
		# have they requested initialization yet?
		self.confirmed = False
		#
		#  Code Methods
		#
		self.codes = {}
		self.codes[ codes.client.ALIAS ] = self.mAlias
		self.codes[ codes.client.CORRECT_PROTO ] = self.mCorrectProto
		self.codes[ codes.client.CHANGED_LINE ] = self.mChangedLine
		self.codes[ codes.client.DISCONNECT ] = self.mDisconnect
		self.codes[ codes.client.FILE_CHAT ] = self.mFileChat
		self.codes[ codes.client.MODIFICATION ] = self.mModification
		self.codes[ codes.client.REQUEST_FILE ] = self.mRequestFile
		self.codes[ codes.client.REQUEST_FILES ] = self.mRequestFiles
		self.codes[ codes.client.REQUEST_LINES ] = self.mRequestLines
		self.codes[ codes.client.REQUEST_USERS ] = self.mRequestUsers
		self.codes[ codes.client.STOP_FILE ] = self.mStopFile
	
	#
	#  Misc Functionality
	#

	def distSendMod( self , fHash , text , everyone = False ):
		"""New function to send a file modification only to those who are viewing/editing that file."""
		for key in protocols.iterkeys():
			if fHash in protocols[ key ].files and ( everyone or self.id != key ):
				protocols[key].sendLine( text )
		return True
	
	def error( self ):
		"""Sends an error message to the client"""
		self.sendLine( codes.ERROR + ":" )
		return True

	def hashID( self , id ):
		"""Returns an appropriately hashed ID"""
		return encode( md5.new( str( id ) ).digest() )

	def sendFile( self , key ):
		"""Splits a file into equal chunks and sends it to the client"""
		# FIXME: if the file is quite large, this could cause a delay on the rest of the server.
		#        in fact, it probably already does, it just isn't very noticable. should we have
		#        the client prompt us for each line, that way this won't bog down the rest of the
		#        traffic? After all, this isn't threaded...
		#        -- Daniel Bickett
		self.sendLine( "%s:%s" % ( codes.server.FILE_START , key ) )
		pieces = len( files[ key ]['data'] ) / 80
		if pieces < 1:
			pieces = 1
		splitFile = splitEqual( files[ key ]['data'] , pieces )
		for slice in splitFile:
			self.sendLine( "%s:%s:%s" % ( codes.server.FILE_LINE , key , encode( slice ) ) )
		self.sendLine( "%s:%s" % ( codes.server.FILE_DONE , key ) )
		return True

	def sendUserLines( self , fHash ):
		for user, line in files[ fHash ]['lines'].iteritems():
			if user != self.id:
				self.sendLine( "%s:%s:%s:%d" % ( codes.client.CHANGED_LINE , user , fHash , line ) )
		return True

	#
	#  Code Methods
	#

	def mAlias( self , args ):
		# this should already be encoded by the client, and we're not decoding it
		# TODO: The reason I'm not announcing this is because it's only meant to be
		#       done in the initiation process, so everyone else will hear about it
		#       when they verify they're running the correct protocol.
		if not hasattr( self , 'alias' ):
			self.alias = args
			return True
		else:
			self.error()
			return False

	def mChangedLine( self , args ):
		# This user has moved his caret to a different line
		fHash, lineNum = args.split(":")
		# make sure this file exists
		if fHash in files:
			try:
				# make sure they sent a proper integer
				lineNum = int( lineNum )
				# update the line table
				files[ fHash ]['lines'][ self.id ] = lineNum
				# tell everyone else
				self.distSendMod( fHash , "%s:%s:%s:%d" % ( codes.client.CHANGED_LINE , self.id , fHash , lineNum ) )
				return True
			except ValueError:
				# if they didn't, yell at them
				self.error()
				return False
		else:
			# if it doesn't, complain
			self.error()
			return False

	def mCorrectProto( self , args ):
		if not self.confirmed and self.alias:
			# This user has confirmed that he is running the correct protocol, so we'll proceed
			# to assert ourselves as here
			protocols[ self.id ] = self
			# announce it to everyone else
			distSend( "%s:%s:%s:%s" % ( codes.server.NEW_USER , self.id , self.ip , self.alias ) , self.id )
			# make sure they can't do this again
			self.confirmed = True
		else:
			# snap at the client
			self.error()

	def mDisconnect( self , args ):
		# This user has informed us that he will be disconnecting
		self.connectionLost()

	def mFileChat( self , args ):
		# A user has said something in the chat of a file
		fHash, statement = args.split(":")
		if fHash in self.files:
			statement = base64.decodestring( statement ).strip()
			if len( statement ) > 0 and len( statement ) <= 175:
				# make sure the text is less than or equal to the max length: 175,
				# and greater than zero, stripped of whitespace
				statement = base64.encodestring( statement ).strip()
				self.distSendMod( fHash , "%s:%s:%s:%s" % ( codes.client.FILE_CHAT , fHash , self.id , statement ) , True )
				return True
			else:
				self.error()
				return False
		else:
			self.error()
			return False

	def mModification( self , args ):
		# This user has modified the document
		mod, fHash, pos, mArg = args.split(":")
		pos = int( pos )
		if mod == codes.client.mod.INSERT:
			text = mArg
			change = strIns( files[ fHash ]['data'] , pos , base64.decodestring( text ) )
			if change:
				files[ fHash ]['data'] = change
				self.distSendMod( fHash , "%s:%s:%s:%s:%d:%s" % ( codes.client.MODIFICATION , mod , self.id , fHash , pos , text ) )
			else:
				self.error()
		elif mod == codes.client.mod.DELETE:
			length = int( mArg )
			change = strDel( files[ fHash ]['data'] , pos , length )
			if change != False:
				files[ fHash ]['data'] = change
				self.distSendMod( fHash , "%s:%s:%s:%s:%d:%d" % ( codes.client.MODIFICATION , mod , self.id , fHash , pos , length ) )
			else:
				self.error()

	def mRequestFile( self , args ):
		# This user has requested a file
		fHash = args.strip()
		if fHash in files:
			# add this user to the user line table
			files[ fHash ]['lines'][ self.id ] = 0
			# add this file to this user's file table
			self.files.append( fHash )
			# tell everyone else running this file that this user has joined the file
			self.distSendMod( fHash , "%s:%s:%s:%s" % ( codes.client.REQUEST_FILE , fHash , self.id , self.alias ) )
			# send this file to this user
			self.sendFile( fHash )
			return True
		else:
			self.error()
			return False

	def mRequestFiles( self , args ):
		# send this client the current files
		for key in files.iterkeys():
			self.sendLine( "%s:%s:%d" % ( codes.server.FILE_CRED , key , len( files[ key ]['data'] ) ) )

	def mRequestLines( self , args ):
		# send this client the line table for the specified file
		fHash = args
		if fHash in self.files:
			self.sendUserLines( fHash )
			return True
		else:
			self.error()
			return False 

	def mRequestUsers( self , args ):
		# send this client the current people
		for protocol in protocols.itervalues():
			self.sendLine( "%s:%s:%s:%s" % ( codes.server.NEW_USER , protocol.id , protocol.ip , protocol.alias ) )

	def mStopFile( self , args ):
		# This user is canceling a file stream
		fHash = args.strip()
		if fHash in self.files and self.id in files[ fHash ]['lines']:
			del files[ fHash ]['lines'][ self.id ]
			self.files.remove( fHash )
			self.distSendMod( fHash , "%s:%s:%s" % ( codes.client.STOP_FILE , fHash , self.id ) )
			return True
		else:
			self.error()
			return False

	#
	#  Overides
	#
	
	def connectionMade( self ):
		# send the version of the protocol this server is running
		self.sendLine( "%s:%s" % ( codes.server.PROTO_VERS , codes.VERS ) )
		# hash our id, encode our ip
		self.id = self.hashID( id( self ) )
		self.ip = encode( self.transport.getPeer().host )

	def connectionLost( self , reason = None ):
		if hasattr( self , 'id' ) and self.id in protocols:
			# Remove this user from all of the files
			for fHash in self.files:
				self.mStopFile( fHash )
			del protocols[ self.id ]
			distSend( "%s:%s" % ( codes.client.DISCONNECT , self.id ) )

	def lineReceived( self , line ):
		line = line.split( ":" , 1 )
		if len( line ) == 2:
			code, arg = line
			if code in self.codes:
				self.codes[ code ]( arg )
			else:
				self.error()

def serveFile( f ):
	if os.path.isfile( f ):
		# load the file into memory
		data = file( f , 'r' ).read()
		# put it in the proper place
		dir, f = os.path.split( f )
		files[ encode( f ) ] = { 'name' : f , 'data' : data , 'lines' : {} }
		return True
	else:
		return False

if __name__ == '__main__':
	import sys
	usage = "Each argument should be a filepath to serve"
	if len( sys.argv ) != 1:
		for f in sys.argv[1:]:
			if serveFile( f ):
				print "File loaded:" , f
			else:
				print "Error loading:" , f
		print "Running server..."
		ConcurrenceFactory = Factory()
		ConcurrenceFactory.protocol = ConcurrenceServerProtocol
		reactor.listenTCP( 3928 , ConcurrenceFactory )
		reactor.run()
	else:
		print usage